use std::cell::RefCell;
use std::collections::HashMap;

use bytes::{BufMut, Bytes, BytesMut};
use resvg::usvg::tiny_skia_path::PathBuilder;
use tiny_skia::{
    BlendMode, Color, FilterQuality, IntSize, LineCap, LineJoin, Paint, Pixmap, PixmapPaint, PremultipliedColorU8, Rect, Shader, Stroke, Transform,
};

use loupedeck_serial::util::Endianness;

use crate::icons::LoadedIcon;
use crate::model::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use crate::model::image_filter::ImageFilterTransform;
use crate::runner::graphics::labels::LabelRenderer;
use crate::runner::state::Key;

#[derive(Debug)]
struct GraphicsContext {
    label_renderer: RefCell<LabelRenderer>,
    buffer_endianness: Endianness,
    global_icon_filter_by_pack_id: HashMap<String, ImageFilterTransform>,
    loaded_icons: HashMap<IconDescriptor, LoadedIcon>,
}

pub fn render_key(context: &GraphicsContext, key_size: IntSize, state: Option<&Key>) -> Bytes {
    let mut pixmap = Pixmap::new(key_size.width(), key_size.height()).unwrap();

    if let Some(state) = state {
        let style = state.style.merge_over(&state.base_style);

        if style.icon.source != IconDescriptorSource::None {
            let loaded_icon = &context.loaded_icons[&state.icon];

            let filter = if let Some(global_filter) = state.icon.source.pack_id().map(|i| context.global_icon_filter_by_pack_id.get(i)).flatten() {
                &state.icon.filter.transform.combine_after(global_filter)
            } else {
                &state.icon.filter.transform
            };

            let scale = filter.scale / loaded_icon.pre_scale;

            let scaled_size = IntSize::from_wh(loaded_icon.pixmap.width(), loaded_icon.pixmap.height())
                .unwrap()
                .scale_by(scale)
                .unwrap();

            static PAINT: PixmapPaint = PixmapPaint {
                opacity: 1.0,
                blend_mode: BlendMode::SourceOver,
                quality: FilterQuality::Bicubic,
            };

            pixmap.draw_pixmap(
                (((key_size.width() as i32 - scaled_size.width() as i32) / 2) as f32 / scale).round() as i32,
                (((key_size.height() as i32 - scaled_size.height() as i32) / 2) as f32 / scale).round() as i32,
                loaded_icon.pixmap.as_ref(),
                &PAINT,
                Transform::from_scale(scale, scale).post_rotate_at(
                    (filter.clockwise_quarter_rotations as f32) * 90.0,
                    key_size.width() as f32 / 2.0,
                    key_size.height() as f32 / 2.0,
                ),
                None,
            );
        }

        if !state.label.is_empty() {
            context.label_renderer.borrow_mut().render(&mut pixmap, &state.label);
        }

        if let Some(color) = state.st {
            let path = PathBuilder::from_rect(Rect::from_xywh(-1.0, -2.0, pixmap.width() as f32, pixmap.height() as f32).unwrap());

            let paint = Paint {
                shader: Shader::SolidColor(Color::from_rgba8(color.r, color.g, color.b, 255)),
                ..Paint::default()
            };

            static STROKE: Stroke = Stroke {
                width: 15.0,
                miter_limit: 4.0,
                line_cap: LineCap::Butt,
                line_join: LineJoin::Round,
                dash: None,
            };

            pixmap.stroke_path(&path, &paint, &STROKE, Transform::identity(), None);
        }
    }

    convert_pixels_to_rgb565(pixmap.pixels(), buffer_endianness).freeze()
}

fn convert_pixels_to_rgb565(pixels: &[PremultipliedColorU8], endianness: Endianness) -> BytesMut {
    let pixel_count = pixels.len();

    let mut result = BytesMut::with_capacity(pixel_count * 2);

    for pixel in pixels {
        let red = pixel.red() as u16;
        let green = pixel.green() as u16;
        let blue = pixel.blue() as u16;
        let color = ((red & 0b11111000) << 8) | ((green & 0b11111100) << 3) | (blue.wrapping_shr(3));

        match endianness {
            Endianness::LittleEndian => result.put_u16_le(color),
            Endianness::BigEndian => result.put_u16(color),
        }
    }

    result
}

pub mod labels {
    use cosmic_text::{Align, Attrs, AttrsList, Buffer, BufferLine, FontSystem, Metrics, Shaping, SwashCache};
    use tiny_skia::{Color, Paint, Pixmap, Rect, Shader, Transform};

    pub struct LabelRenderer {
        font_system: FontSystem,
        swash_cache: SwashCache,
        buffer: Buffer,
    }

    impl LabelRenderer {
        pub fn new() -> Self {
            let mut font_system = FontSystem::new();
            let buffer = Buffer::new(&mut font_system, Metrics::new(11.0, 11.0));

            LabelRenderer {
                font_system,
                swash_cache: SwashCache::new(),
                buffer,
            }
        }

        pub fn render(&mut self, pixmap: &mut Pixmap, text: &String) {
            let attrs = Attrs::new();
            let mut line = BufferLine::new(text, AttrsList::new(attrs), Shaping::Advanced);
            line.set_align(Some(Align::Center));

            self.buffer.lines.clear();
            self.buffer.lines.push(line);

            const PADDING: f32 = 12.0;
            self.buffer.set_size(
                &mut self.font_system,
                pixmap.width() as f32 - PADDING * 2.0,
                pixmap.height() as f32 - PADDING * 2.0,
            );

            self.buffer.shape_until_scroll(&mut self.font_system);

            self.buffer.draw(
                &mut self.font_system,
                &mut self.swash_cache,
                cosmic_text::Color::rgb(255, 255, 255),
                |x, y, w, h, color| {
                    pixmap.fill_rect(
                        Rect::from_xywh(x as f32 + PADDING, y as f32 + PADDING, w as f32, h as f32).unwrap(),
                        &Paint {
                            shader: Shader::SolidColor(Color::from_rgba8(color.r(), color.g(), color.b(), color.a())),
                            ..Paint::default()
                        },
                        Transform::identity(),
                        None,
                    );
                },
            );
        }
    }
}
